/*-
 * Copyright @BABOLO (C)2002,2003 http://www.babolo.ru/
 * All rights reserved.
 * Based on:
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright @BABOLO (C)2002,2003 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: jail.c,v 1.5 2008/12/22 05:27:50 babolo Exp $";
#endif /* not lint */

#include <sys/time.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/resource.h>
#include <sys/rtprio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <login_cap.h>
#include <sys/jail.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <pwd.h>
#include <grp.h>
#include <babolo/BLINflag.h>

#define J_PATH argv[0]
#define J_HOSTNAME argv[1]
#define J_HOSTADDR argv[2]
#define J_EXEC_N 3
#define J_EXEC argv[J_EXEC_N]

#define JAILUP_NICE  0x000100 /*   nice */
#define JAILUP_GROUP 0x000200 /*    */
#define JAILUP_CLASS 0x000400 /*   login class */
#define JAILUP_LOGIN 0x001000 /*  setusercontext() */
#define JAILUP_UNAME 0x002000 /*  uname */
#define JAILUP_UID   0x004000 /*  uid */

void
usage(int ex) {
    fprintf( stderr
           , "%s @BABOLO V.M "VERS"  "DATE" from jailup(8) package\n"
             "Usage:\n"
             "%s [-l ][-u user ][-g group ][-n nice ][-p pidfile ]...[-P pidfile ] path hostname ip-number command \n"
             "   -l for use login classes database in jailed environment for setusercontext()\n"
             "   -p write pid to pidfile before jailed, repetitive\n"
             "   -P write pid to pidfile after jailed, only last -P used\n"
           , getprogname()
           , getprogname()
           );
    exit(ex);
}

int
main(int argc, char **argv) {
    char *gname = NULL, *uname = NULL, *cnice = NULL, *pidfl = NULL, *flpid = NULL;
    struct passwd *chuserp = NULL;
    struct group  *chgrpp = NULL;
    struct in_addr in;
    struct rtprio  rtp;
    struct jail    j;
    unsigned int ufl;
    u_int32_t    flags = BLIN_VER1;
    int          i, nice = 0;
    login_cap_t *lc = NULL;
    uid_t chuser;
    gid_t tgrp;
    char *end;
    FILE *pid;

    while ((i = getopt(argc, argv, "g:ln:p:P:qu:v?")) != -1) {
        switch (i) {
        case 'g': gname = optarg;            break;
        case 'l': flags |= JAILUP_LOGIN;     break;
        case 'n': cnice = optarg;            break;
        case 'p': flpid = optarg;            break;
        case 'P': pidfl = optarg;            break;
        case 'q': flags &= ~BLIN_VERMASK;    break;
        case 'u': uname = optarg;            break;
        case 'v': BLIN_VERBOSE(flags);       break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }

    argc -= optind;
    if  (argc < 4) {
        ifBLIN_QV1(flags) usage(EX_USAGE);
        exit(EX_USAGE);
    }
    argv += optind;
    if  (cnice) {
        if  (*cnice == '=') {
            flags |= JAILUP_LOGIN;
            ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: LOGIN for nice\n");
        } else if (*cnice > '9') {
            flags |= JAILUP_CLASS | JAILUP_LOGIN;
            ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: login class = %s\n", cnice);
        } else {
            flags |= JAILUP_NICE;
            errno = 0;
            nice = strtol(optarg, &end, 0);
            if  (errno) {
                ifBLIN_QV1(flags) warn("Invalid nice value %s", optarg);
                exit(EX_DATAERR);
            } else if (*end) {
                ifBLIN_QV1(flags) warnx("Invalid nice value %s", optarg);
                exit(EX_DATAERR);
            }
            if  (nice > PRIO_MAX) {
	        rtp.type = RTP_PRIO_IDLE;
	        nice = nice - PRIO_MAX - 1;
	        rtp.prio = (nice > RTP_PRIO_MAX) ? RTP_PRIO_MAX : nice;
                ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: idle prio %d\n", rtp.prio);
	        if  (rtprio(RTP_SET, 0, &rtp) < 0) {
                    ifBLIN_QV2(flags) warn("rtprio idle %d", rtp.prio);
                }
            } else if (nice < PRIO_MIN) {
	        rtp.type = RTP_PRIO_REALTIME;
	        nice = nice - PRIO_MIN + RTP_PRIO_MAX + 1;
	        rtp.prio = (nice < 0) ? 0 : nice;
                ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: rt prio %d\n", rtp.prio);
	        if  (rtprio(RTP_SET, 0, &rtp) < 0) {
                    ifBLIN_QV2(flags) warn("rtprio realtime %d", rtp.prio);
                }
            } else {
                ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: nice %d\n", nice);
                if  (setpriority(PRIO_PROCESS, 0, (int)nice) < 0) {
                    ifBLIN_QV2(flags) warn("setpriority %d", nice);
    }   }   }   }
    if  (chdir(J_PATH)) {
        ifBLIN_QV1(flags) warn("chdir %s", J_PATH);
        exit(EX_UNAVAILABLE);
    }
    memset(&j, 0, sizeof(j));
    j.version = 0;
    j.path = J_PATH;
    j.hostname = J_HOSTNAME;
    if  (!inet_aton(J_HOSTADDR, &in)) {
        ifBLIN_QV1(flags) warnx("Couldn't make sense of IP %s", J_HOSTADDR);
        exit(EX_NOHOST);
    }
    j.ip_number = ntohl(in.s_addr);
    if  (flpid) {
        pid = fopen(flpid, "w");
        if  (!pid) {
            ifBLIN_QV1(flags) warn("PID before: fopen %s", flpid);
            exit(EX_CANTCREAT);
        }
        fprintf(pid, "%d\n", getpid());
        fclose(pid);
    }
    if  (jail(&j) < 0) {
        ifBLIN_QV1(flags) warn("Imprisonment failed");
        exit(EX_SOFTWARE);
    }
    if  (!gname) {
        tgrp = getgid();
    } else if (*gname == '=') {
        flags |= JAILUP_LOGIN;
        ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: LOGIN for group\n");
    } else {
        flags |= JAILUP_GROUP;
        errno = 0;
        if  ((chgrpp = getgrnam(gname)) == NULL) {
            if  (errno) ifBLIN_QV2(flags) warn("getgrnam %s", uname);
            /* Try if group is by number, not by name */
            tgrp = (gid_t)strtoul(gname, &end, 0);
            if  (*end) {
                ifBLIN_QV1(flags) warnx("No such group %s", gname);
                exit(EX_DATAERR);
            }
            ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: gid %d\n", tgrp);
        } else {
            tgrp = chgrpp->gr_gid;
            ifBLIN_QV3(flags)
                fprintf(stderr, "jailup: jail: gname %s gid %d\n", chgrpp->gr_name, tgrp);
        }
        if  (setgid(tgrp) < 0) {
            ifBLIN_QV1(flags) warn("setgid %d", tgrp);;
            exit(EX_NOPERM);
    }   }
    if  (!uname) {
        chuser = getuid();
    } else {
        errno = 0;
        if  (!(chuserp = getpwnam(uname))) {
            if  (errno) ifBLIN_QV2(flags) warn("getpwnam %s", uname);
            /* Try if user is by number, not by name */
            chuser = (uid_t)strtoul(uname, &end, 10);
            if  (*end) {
                ifBLIN_QV1(flags) warnx("No such user %s", uname);
                exit(EX_NOUSER);
            }
            errno = 0;
            if  ((chuserp = getpwuid(chuser))) {
                ifBLIN_QV3(flags) fprintf( stderr, "jailup: jail: uid %d, uname %s\n"
                                         , chuserp->pw_uid
                                         , chuserp->pw_name
                                         );
            } else {
                if  (errno) ifBLIN_QV2(flags) warn("getpwuid %d", chuser);
                ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: uid %d\n", chuser);
            }
        } else {
            flags |= JAILUP_UNAME;
            chuser = chuserp->pw_uid;
            ifBLIN_QV3(flags)
                fprintf(stderr, "jailup: jail: uname %s, uid %d\n", chuserp->pw_name, chuserp->pw_uid);
        }
        flags |= JAILUP_UID;
    }
    if  ((flags & JAILUP_CLASS) && !(lc = login_getclass(cnice))) {
        ifBLIN_QV1(flags) warnx("No class %s", cnice);
    }
    if  (flags & (JAILUP_LOGIN)) {
        ufl = LOGIN_SETALL;
        if  (setsid() == -1) {
            ifBLIN_QV1(flags) warn("setsid");
            exit(EX_NOPERM);
        }
        if  (flags & JAILUP_NICE) ufl &= ~LOGIN_SETPRIORITY;
        if  (flags & JAILUP_GROUP) ufl &= ~LOGIN_SETGROUP;
        if  (!uname) ufl &= ~LOGIN_SETUSER;
        ifBLIN_QV3(flags) fprintf( stderr, "jailup: jail: setusercontext(%s, %s, %d, %08X)\n"
                                 , lc ? "lc" : ""
                                 , chuserp ? "chuserp" : ""
                                 , chuser, ufl
                                 );
        if  (setusercontext(lc, chuserp, chuser, ufl) < 0) {
            ifBLIN_QV1(flags) warn("setusercontext");
            exit(EX_SOFTWARE);
        }
    } else {
        if  (flags & JAILUP_UNAME) {
            ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: setlogin %s\n", chuserp->pw_name);
            if  (setlogin(chuserp->pw_name) < 0) {
                ifBLIN_QV2(flags) warn("setlogin %s", chuserp->pw_name);
        }   }
        if  (flags & JAILUP_UID) {
            ifBLIN_QV3(flags) fprintf(stderr, "jailup: jail: setuid %d\n", chuser);
            if  (setuid(chuser) < 0) {
                ifBLIN_QV2(flags) warn("setuid %d", chuser);
                exit(EX_NOPERM);
    }   }   }
    login_close(lc);
    if  (pidfl) {
        pid = fopen(pidfl, "w");
        if  (!pid) {
            ifBLIN_QV1(flags) warn("PID after: fopen %s", pidfl);
            exit(EX_CANTCREAT);
        }
        fprintf(pid, "%d\n", getpid());
        fclose(pid);
    }
    ifBLIN_QV3(flags) {
        for (i = 0; (argv + J_EXEC_N)[i]; i++)
            fprintf(stderr, "jailup: jail: exec[%d]=%s~\n", i, (argv + J_EXEC_N)[i]);
    }
    execv(J_EXEC, argv + J_EXEC_N);
    ifBLIN_QV1(flags) warn("execv %s", J_EXEC);
    exit(EX_UNAVAILABLE);
}
